#!/usr/bin/env python

'''Assetto Corsa Websockets Server.'''

import argparse
import asyncio
import logging
import sys

from ac_websocket_server.constants import HOST, PORT, PROG
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.server import WebsocketsServer


def get_options(args):
    '''Parse arguments for options.'''
    # pylint: disable=line-too-long

    parser = argparse.ArgumentParser(prog=PROG,
                                     description='''
                                     A web-sockets server to control a local Assetto Corsa Server.''')

    parser.add_argument("--game", type=str,
                        help="acserver root directory")

    parser.add_argument("--host", type=str, default=HOST,
                        help="host to connect/listen")

    parser.add_argument("--port", type=int, default=PORT,
                        help="port to connect/listen")

    parser.add_argument("--verbose", action="store_true",
                        help="show verbose output")

    return parser.parse_args(args)


async def main(args):
    '''Command line main'''

    options = get_options(args)

    logger = logging.getLogger('ac-ws')

    if options.verbose:
        logger.setLevel(logging.DEBUG)

    try:
        server = WebsocketsServer(host=options.host,
                                  port=options.port,
                                  server_directory=options.game)

        await server.start()

    except WebsocketsServerError as err:
        print(f'ERROR: {str(err)}')


if __name__ == '__main__':

    asyncio.run(main(sys.argv[1:]))



'''Assetto Corsa Websocket Server Class'''

import asyncio
import logging
import os
from typing import List
import websockets

from ac_websocket_server.connection import WebSocketConnection
from ac_websocket_server.debug import monitor_tasks
from ac_websocket_server.constants import HOST, PORT
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.game import GameServer
from ac_websocket_server.grid import Grid
from ac_websocket_server.handlers import handler
from ac_websocket_server.observer import Observer
from ac_websocket_server.protocol import Protocol
from ac_websocket_server.tracker import TrackerServer

EXTRA_DEBUG = False


class WebsocketsServer():
    '''Represents an Assetto Corsa WebSocket Server.

    Allows control of an Assetto Corsa server with a websockets interface.'''
    # pylint: disable=logging-fstring-interpolation

    def __init__(self,
                 server_directory: str = None,
                 host: str = HOST,
                 port: int = PORT
                 ) -> None:

        self._logger = logging.getLogger('ac-ws.ws-server')

        if EXTRA_DEBUG:
            asyncio.get_event_loop().create_task(monitor_tasks())

        self._connections: List[WebSocketConnection] = []

        self.host = host
        self.port = port

        if not server_directory:
            self.server_directory = os.getcwd()
        else:
            self.server_directory = server_directory

        try:
            self.game = GameServer(server_directory=self.server_directory)
        except WebsocketsServerError as error:
            self._logger.error(f'Fatal error {error}')
            raise

        try:
            self.tracker = TrackerServer(f'{self.server_directory}/stracker')
        except WebsocketsServerError as error:
            self._logger.error(f'Unable to start tracker: {error}')

        self.stop_server: asyncio.Future = None

    async def broadcast(self, msg: str):
        '''Broadcast message to all connected websockets.'''

        for connection in self._connections:
            await connection.put(msg)

    async def handler(self, websocket):
        '''ACWS handler function for websocket connection'''

        connection = WebSocketConnection(server=self, websocket=websocket)

        self._connections.append(connection)

        await connection.send(Protocol.success(
            msg=f'Welcome to the Assetto Corsa WebSocket server running at {self.host}:{self.port}'))

        await handler(websocket, connection.consumer, connection.producer)

        connection.close()

    async def start(self):
        '''Start the websocket server'''

        try:

            self._logger.info('Starting websocket server')

            self.stop_server = asyncio.Future()

            async with websockets.serve(self.handler, self.host, self.port):
                await self.stop_server

            self._logger.info('Stopping websocket server')

        except KeyboardInterrupt:
            self._logger.info('Interupting the server')

    async def stop(self):
        '''Stop the websocket server'''

        self.stop_server.set_result(True)

    async def shutdown(self, message: str = None):
        '''
        Shutdown the ACWS server.

        Note that running AC servers and trackers will NOT be stopped.
        '''

        await self.broadcast(Protocol.success(
            msg=f'Shutting down the WebSocket server running at {self.host}:{self.port}'))

        self._logger.info('Shutting down the server')
        await self.stop()



'''Assetto Corsa Websocket handlers'''

import asyncio
import websockets

from ac_websocket_server.error import WebsocketsServerError


async def consumer_handler(websocket, consumer):
    '''Consumer messages from websocket and pass to user function'''
    try:
        async for message in websocket:
            await consumer(message)
    except websockets.exceptions.ConnectionClosed as error:
        print(error)
        raise WebsocketsServerError(error) from error
    except Exception as error:
        print(error)
        raise


async def handler(websocket, consumer, producer):
    '''Setup consumer and producer handlers.'''

    consumer_task = asyncio.create_task(consumer_handler(websocket, consumer))
    producer_task = asyncio.create_task(producer_handler(websocket, producer))
    _done, pending = await asyncio.wait(
        [consumer_task, producer_task],
        return_when=asyncio.FIRST_COMPLETED,
    )
    for task in pending:
        task.cancel()


async def producer_handler(websocket, producer):
    '''Send user producer output to websocket'''
    # pylint: disable=broad-except
    while True:
        try:
            await websocket.send(await producer())
        except websockets.exceptions.ConnectionClosed as error:
            print(error)
            raise WebsocketsServerError(error) from error
        except Exception as error:
            print(error)
            raise



'''Assetto Corsa Game Server Class'''

import asyncio
import configparser
import hashlib
import json
import logging
import os
import sys
from dataclasses import dataclass, field
from datetime import datetime
import subprocess
from typing import Dict, List

from ac_websocket_server.entries import EntryList
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.grid import Grid
from ac_websocket_server.objects import (DriverInfo, EnhancedJSONEncoder,
                                         EntryInfo, SessionInfo)
from ac_websocket_server.observer import ObserverNotifier
from ac_websocket_server.protocol import Protocol
from ac_websocket_server.watcher import Watcher


@dataclass
class GameServer(ObserverNotifier):
    '''Represents an Assetto Corsa Server.'''
    # pylint: disable=logging-fstring-interpolation, invalid-name

    server_directory: str

    version: str = field(default='n/a', init=False)
    timestamp: str = field(default='n/a', init=False)
    name: str = field(init=False)
    track: str = field(init=False)
    cars: str = field(init=False)
    http_port: int = field(init=False)
    tcp_port: int = field(init=False)
    udp_port: int = field(init=False)
    drivers: Dict[str, DriverInfo] = field(init=False)
    entries: Dict[int, EntryInfo] = field(init=False)
    sessions: Dict[str, SessionInfo] = field(init=False)

    def __post_init__(self):

        super().__init__()

        self.__logger = logging.getLogger('ac-ws.game')

        self.drivers = {}
        self.entries = {}
        self.sessions = {}

        if os.path.exists(f'{self.server_directory}/acServer.py'):
            self.__cwd = None
            self.__exe = f'{self.server_directory}/acServer.py'
            self.__hash = None
        else:
            self.__cwd = self.server_directory
            if sys.platform == 'linux':
                self.__exe = f'{self.server_directory}/acServer'
                self.__hash = 'f781ddfe02e68adfa170b28d0eccbbdc'
            else:
                self.__exe = f'{self.server_directory}/acServer.exe'
                self.__hash = '357e1f1fd8451eac2d567e154f5ee537'

        self.parse_server_cfg()
        self.parse_entry_list()

        self.grid = Grid(server_directory=self.server_directory,
                         track=self.track,
                         entry_list=self.entry_list)

        self.__logfile_stdout: str
        self.__logfile_stderr: str

        self.__watcher_stdout: Watcher

        self.__process: asyncio.subprocess.Process

        self.running = False

    async def consumer(self, message_words: List[str]):
        '''Consume args destined for the server'''

        message_funcs = {'drivers': self.__drivers,
                         'entries': self.__entries,
                         'info': self.__info,
                         'sessions': self.__sessions,
                         'start': self.start,
                         'stop': self.stop,
                         'restart': self.restart}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]]()

    async def __drivers(self):
        '''Show game drivers info as part of a JSON reply'''
        await self.put(Protocol.success({'drivers': self.drivers}))

    async def __entries(self):
        '''Show game entries info as part of a JSON reply'''
        await self.put(Protocol.success({'entries': self.entries}))

    async def __info(self):
        '''Show game server info as a JSON string'''
        await self.put(Protocol.success({'server': self}))

    async def notify(self, notifier):
        '''Receive a notification of a new message from log watcher.'''

        message = await notifier.get(self)

        try:
            item = json.loads(message)

            if item['type'] == 'ServerInfo':
                self.version = item['body']['version']
                self.timestamp = item['body']['timestamp']
                self.track = item['body']['track']
                self.cars = item['body']['cars']

                await self.put(Protocol.success({'server': item['body']}))

            if item['type'] == 'DriverInfo' and item['body']['msg'] == 'joining':

                body = item['body']
                name = body['name']

                driver_info = DriverInfo()

                driver_info.name = name
                driver_info.host = body['host']
                driver_info.port = body['port']
                driver_info.car = body['car']
                driver_info.guid = body['guid']
                driver_info.ballast = body['ballast']
                driver_info.msg = body['msg']

                self.drivers[driver_info.name] = driver_info

                self.__logger.debug(f'Driver {name} joining')
                await self.put(Protocol.success({'driver': item['body']}))

            if item['type'] == 'DriverInfo' and item['body']['msg'] == 'leaving':
                body = item['body']
                name = body['name']
                del self.drivers[name]

                self.__logger.debug(f'Driver {name} leaving')
                await self.put(Protocol.success({'driver': item['body']}))

            if item['type'] == 'SessionInfo':

                body = item['body']
                session_type = body['type']

                session_info = SessionInfo()

                session_info.type = session_type
                session_info.laps = body['laps']
                session_info.time = body['time']

                self.sessions[session_type] = session_info

                await self.put(Protocol.success({'sessions': session_info}))

        except json.JSONDecodeError:
            pass

        # await self.put(message)

    def parse_entry_list(self):
        '''Parse entry list file and update attributes'''

        entry_list_file_name = f'{self.server_directory}/cfg/entry_list.ini'

        if not os.path.exists(entry_list_file_name):
            error_message = f'Missing entry_list.ini file in {self.server_directory}'
            self.__logger.error(error_message)
            raise WebsocketsServerError(error_message)

        self.entry_list = EntryList(entry_list_file_name)
        self.entries = self.entry_list.entries

    def parse_server_cfg(self):
        '''Parse server config file and update attributes'''

        server_cfg_file_name = f'{self.server_directory}/cfg/server_cfg.ini'

        if not os.path.exists(server_cfg_file_name):
            error_message = f'Missing server_cfg.ini file in {self.server_directory}'
            self.__logger.error(error_message)
            raise WebsocketsServerError(error_message)

        cfg = configparser.ConfigParser()
        try:
            cfg.read(server_cfg_file_name)

            self.name = cfg['SERVER']['NAME']
            self.cars = cfg['SERVER']['CARS']
            self.track = cfg['SERVER']['TRACK']
            self.http_port = cfg['SERVER']['HTTP_PORT']
            self.tcp_port = cfg['SERVER']['TCP_PORT']
            self.udp_port = cfg['SERVER']['UDP_PORT']

            for session in ['PRACTICE', 'QUALIFY', 'RACE']:
                if cfg.has_section(session):
                    name = cfg[session].get('NAME')
                    time = cfg[session].get('TIME', 0)
                    laps = cfg[session].get('LAPS', 0)
                    self.sessions[name] = SessionInfo(
                        name, laps=laps, time=time)

        except configparser.Error as e:
            raise WebsocketsServerError(e) from e

    async def restart(self):
        ''''Re-start the game server'''
        self.__logger.info('Re-starting game server')
        await self.stop()
        await self.start()

    async def __sessions(self):
        '''Show game sessions info as a JSON string'''
        await self.put(Protocol.success({'sessions': self.sessions}))

    async def start(self):
        '''Start the game server.'''

        # Parse and update status before start (for re-start post __init__)
        try:
            self.parse_server_cfg()
            self.parse_entry_list()
            self.grid.track = self.track
        except WebsocketsServerError as e:
            self.__logger.error('Command failed - configuration error')
            await self.put(Protocol.error(
                msg=f'start command failed - configuration error: {e}'))
            return

        if self.running:
            self.__logger.info('Command ignored - already running')
            await self.put(Protocol.error(
                msg='start command ignored - game already running'))
            return

        if self.__hash:
            try:
                with open(self.__exe, 'rb') as file_to_check:
                    data = file_to_check.read()
                    if self.__hash != hashlib.md5(data).hexdigest():
                        raise WebsocketsServerError(
                            f'{self.__exe} checksum mismatch')
            except FileNotFoundError as e:
                raise WebsocketsServerError(
                    f'{self.__exe} missing') from e

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

        self.__logger.info('Starting game server')

        os.makedirs(f'{self.server_directory}/logs/session', exist_ok=True)
        os.makedirs(f'{self.server_directory}/logs/error', exist_ok=True)

        self.__logfile_stdout = f'{self.server_directory}/logs/session/output{timestamp}.log'
        self.__logfile_stderr = f'{self.server_directory}/logs/error/error{timestamp}.log'

        session_file = open(self.__logfile_stdout, 'w', encoding='utf-8')
        error_file = open(self.__logfile_stderr, 'w', encoding='utf-8')

        # Set platform specific options to start process in seperate process group
        kwargs = {}
        if sys.platform.startswith('win'):
            kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP | subprocess.DETACHED_PROCESS | subprocess.CREATE_BREAKAWAY_FROM_JOB
        else:
            kwargs['start_new_session'] = True

        try:
            self.__process = await asyncio.create_subprocess_exec(
                self.__exe, cwd=self.__cwd, stdout=session_file, stderr=error_file, **kwargs)

            self.running = True

            self.__logger.info(f'Process pid is: {self.__process.pid}')
            await self.put(Protocol.success(
                msg='Assetto Corsa server started'))

            self.__watcher_stdout = Watcher(self.__logfile_stdout)
            self.__watcher_stdout.subscribe(self)
            await self.__watcher_stdout.start()

        except PermissionError as e:
            self.__logger.error(f'Process did not start: {e}')
            await self.put(Protocol.error(
                msg='Assetto Corsa server did not start'))
            raise WebsocketsServerError(e) from e

    async def stop(self):
        '''Stop the game server'''

        if not self.running:
            self.__logger.info('Command ignored - not running')
            await self.put(Protocol.error(
                msg='stop command ignored - game not running'))
            return

        self.__logger.info('Stopping game server')

        self.__process.terminate()

        status_code = await asyncio.wait_for(self.__process.wait(), None)

        self.__logger.info(f'Game server exited with {status_code}')
        await self.put(Protocol.success(
            msg='Assetto Corsa server stopped'))

        await self.__watcher_stdout.stop()
        self.__watcher_stdout.unsubscribe(self)

        self.running = False



'''Assetto Corsa stracker class'''

import asyncio
import configparser
from dataclasses import dataclass, field
import os
from typing import List

from ac_websocket_server.child import ChildServer
from ac_websocket_server.error import WebsocketsServerError


@dataclass
class TrackerServer(ChildServer):
    '''Represents an stracker server'''

    child_ini_file: str = field(default='stracker.ini')
    child_title: str = field(default='Stracker')
    child_short: str = field(default='stracker')
    is_optional: bool = True

    ac_server_address: str = field(init=False)
    ac_server_cfg_ini: str = field(init=False)
    log_file: str = field(init=False)
    listening_port: int = field(init=False)
    server_name: str = field(init=False)

    async def consumer(self, message_words: List[str]):
        '''Consume args destined for the server'''

        message_funcs = {'start': self.start,
                         'stop': self.stop,
                         'restart': self.restart,
                         'status': self.status}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]]()

    def __post_init__(self):

        super().__post_init__()

        self._exe = f'{self.directory}/stracker.exe'
        self._args = ('--stracker_ini', self.child_ini_file)
        self._hash = '4c288fe128bae5250c51e07e2a510ccd'

    def pre_start_hook(self):
        '''Read (and correct) the stracker.ini file'''

        super().pre_start_hook()
        
        ini_file_name = f'{self.directory}/{self.child_ini_file}'

        if not os.path.exists(ini_file_name):
            msg = f'Missing {self.child_ini_file} file in {self.directory}'
            self._logger.error(msg)
            raise WebsocketsServerError(msg)

        cfg = configparser.ConfigParser()
        try:
            cfg.read(ini_file_name)

            self.ac_server_address = cfg['STRACKER_CONFIG']['ac_server_address']
            self.ac_server_cfg_ini = cfg['STRACKER_CONFIG']['ac_server_cfg_ini']
            self.log_file = cfg['STRACKER_CONFIG']['log_file']
            self.listening_port = cfg['STRACKER_CONFIG']['listening_port']
            self.server_name = cfg['STRACKER_CONFIG']['server_name']

        except configparser.Error as error:
            raise WebsocketsServerError(error) from error




'''Asyncio Debug Utilities.'''

import asyncio

DELAY = 10


async def monitor_tasks():
    '''
    Monitor tasks and print out every DELAY seconds.

    Call with:
    loop.create_task(monitor_tasks())
    '''
    # pylint: disable=expression-not-assigned

    while True:
        print('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>')
        tasks = [
            t for t in asyncio.all_tasks()
            if t is not asyncio.current_task()
        ]
        [t.print_stack(limit=5) for t in tasks]
        await asyncio.sleep(DELAY)



'''Assetto Corsa Grid Class'''

import glob
import logging
import os
from typing import List

from ac_websocket_server.entries import EntryList
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.observer import Notifier
from ac_websocket_server.protocol import Protocol


class Grid(Notifier):
    '''Represents grid setup'''

    def __init__(self, server_directory: str, track: str = None, entry_list: EntryList = None):

        self.__logger = logging.getLogger('ac-ws.grid')

        self.server_directory = server_directory
        self.entry_list = entry_list
        self.track = track

        super().__init__()

    async def consumer(self, message_words: List[str]):
        '''Consume args destined for the server'''

        message_funcs = {'finish': self.__update_finishing,
                         'reverse': self.__update_reversed,
                         'entries': self.__entries,
                         'order': self.__order,
                         'save': self.__save}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]]()

    async def __entries(self):
        '''Show original grid order info as a JSON string'''
        await self.put(Protocol.success({'entries': self.entry_list.show_entries()}))

    async def __order(self):
        '''Show current grid order info as a JSON string'''
        await self.put(Protocol.success({'grid': self.entry_list.show_grid()}))

    async def __save(self):
        '''Write updated entry_list.ini file'''
        try:
            self.entry_list.write(
                f'{self.server_directory}/cfg/entry_list.ini')
            await self.put(Protocol.success(msg='entry_list.ini file update SUCCESS'))
        except WebsocketsServerError as error:
            await self.put(Protocol.error(
                msg=f'entry_list.ini file update FAILURE: {error}'))

    async def update(self, reverse: bool = False):
        '''Update grid based on latest results.JSON file and using sorting rule'''

        if reverse:
            self.entry_list.set_reversed_order()
        else:
            self.entry_list.set_standard_order()

        try:
            result_files = glob.glob(
                os.path.join(f'{self.server_directory}/results', "*RACE.json"))
            if len(result_files) > 0:
                result_file = max(result_files, key=os.path.getctime)
                self.entry_list.parse_result_file(
                    result_file, track=self.track)
            else:
                await self.put(Protocol.error(
                    msg='Unable to parse results file - no file exists'))
                return
        except WebsocketsServerError as error:
            await self.put(Protocol.error(
                msg=f'{result_file} parse FAILURE: {error}'))
            return

        await self.put(Protocol.success(
            msg=f'{result_file} parse SUCCESS'))

        await self.__order()

    async def __update_finishing(self):
        '''Update grid in standard finishing order'''
        await self.update()

    async def __update_reversed(self):
        '''Update grid in reverse finishing order'''
        await self.update(reverse=True)



"""A setuptools based setup module."""

import pathlib
from setuptools import setup, find_packages

from ac_websocket_server.constants import PROG, VERSION


# The text of the README file
README = (pathlib.Path(__file__).parent / "README.md").read_text()

setup(
    name=PROG,
    version=VERSION,
    description='Assetto Corsa Websockets Server',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Mark Hannon',
    author_email='mark.hannon@gmail.com',
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=['aiofiles', 'websockets'],
)



'''Assetto Corsa WebSockets Server Messages'''

from dataclasses import dataclass, is_dataclass, asdict
from enum import Enum
import json
from typing import Union


class EnhancedJSONEncoder(json.JSONEncoder):
    '''
    Enhanced JSON encoder.

    Use as per:
    json.dumps(foo, cls=EnhancedJSONEncoder)

    https://stackoverflow.com/questions/51286748/make-the-python-json-encoder-support-pythons-new-dataclasses
    https://stackoverflow.com/questions/24481852/serialising-an-enum-member-to-json
    '''

    def default(self, o):
        if is_dataclass(o):
            return asdict(o)
        if isinstance(o, Enum):
            return o.value
        return super().default(o)


@dataclass
class DriverInfo:
    '''Represents a driver.'''
    name: str = ''
    host: str = ''
    port: int = 0
    car: str = ''
    guid: str = ''
    ballast: int = 0
    msg: str = ''


@dataclass(repr=False)
class EntryInfo:
    '''Represents an entry in the entry_list.ini file'''
    car_id: str
    model: str = ''
    skin: str = ''
    spectator_mode: int = 0
    drivername: str = ''
    team: str = ''
    guid: str = ''
    ballast: int = 0
    restrictor: int = 0

    def __post_init__(self):
        self.position = 0

    def __repr__(self):
        '''Override __repr__ to match format of entry_list.ini'''

        result = "[" + self.car_id + "]\n"
        result += "MODEL=" + self.model + "\n"
        result += "SKIN=" + self.skin + "\n"
        result += "SPECTATOR_MODE=" + str(self.spectator_mode) + "\n"
        result += "DRIVERNAME=" + self.drivername + "\n"
        result += "TEAM=" + self.team + "\n"
        result += "GUID=" + self.guid + "\n"
        result += "BALLAST=" + str(self.ballast) + "\n"
        result += "RESTRICTOR=" + str(self.restrictor) + "\n"
        result += "\n"

        return result


@dataclass
class ServerInfo:
    '''Represents version information for a server.'''

    version: str = ''
    timestamp: str = ''
    track: str = ''
    cars: str = ''
    msg: str = ''


@dataclass
class SessionInfo:
    '''Represents an individual session in the AC game server'''

    type: str = ''
    laps: int = 0
    time: int = 0
    msg: str = ''


class MessageType(Enum):
    '''Allowable message types'''

    DRIVER_INFO = 'DriverInfo'
    SERVER_INFO = 'ServerInfo'
    SESSION_INFO = 'SessionInfo'


MessageBody = Union[DriverInfo, ServerInfo, SessionInfo]


@dataclass
class Message:
    '''Basic message structure'''
    type: MessageType
    body: MessageBody
